/**
 * \file pnqueue.c
 * 2004-10-21
 */
/*
 *                                     Diagram of Concept
 *
 *
 *                get <--,                                                   ,-- put
 *                       |       < front >        ...       < rear >         |
 *    ,-------------,    |    ,-------------,             ,-------------,    |    ,-------------,
 * ,->| head        |<-. | ,->| node-1      |<-.       ,->| node-n      |<-. | ,->| tail        |
 * |  |-------------|  | | |  |-------------|  |       |  |-------------|  | | |  |-------------|
 * |  | data = NULL |  |   |  | data = ...  |  |  ...  |  | data = ...  |  | * |  | data = NULL |
 * |  |-------------|  |   |  |-------------|  |       |  |-------------|  |   |  |-------------|
 * |  | prev | next |--+---'  | prev | next |--+-     -'  | prev | next |--+---'  | prev | next |
 * |  `-------------'  |      `-------------'  |          `-------------'  |      `-------------'
 * |     |             |         |             |  ...        |             |         | 
 * `-----'             `---------'             `-     -------'             `---------'
 *
 *  put : Put a node between rear and tail.
 *  get : Get a node from front.
 *
 *  - Not use head and tail as a data node.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <pnqueue.h>

#ifndef FAIL
#define	FAIL		(-1)
#endif
#ifndef SUCC
#define	SUCC		0
#endif


void pnqueue_datafree(pnqueue_data_t *data)
{
	if(data->dp) free(data->dp);
}

int pnqueue_init(pnqueue_handle_t *que)
{
	/* allocation */
	que->head = (pnqueue_node_t *)malloc(sizeof(pnqueue_node_t));
	if(!que->head) return FAIL;
	que->tail = (pnqueue_node_t *)malloc(sizeof(pnqueue_node_t));
	if(!que->tail) return FAIL;

	/* initialize head */
	que->head->prev = que->head;
	que->head->next = que->tail;
	que->head->data = NULL;

	/* initialize tail */
	que->tail->prev = que->head;
	que->tail->next = que->tail;
	que->tail->data = NULL;

	return SUCC;
}

void pnqueue_destroy(pnqueue_handle_t *handle)
{
	if(handle->head) free(handle->head);
	if(handle->tail) free(handle->tail);
}

int pnqueue_put(pnqueue_handle_t *handle, pnqueue_data_t *data)
{
	pnqueue_node_t *head=NULL;
	pnqueue_node_t *tail=NULL;
	pnqueue_node_t *new_node=NULL;

	/* mapping temporarily */
	head = handle->head;
	tail = handle->tail;

	/* new node allocation */
	new_node = (pnqueue_node_t *)malloc(sizeof(pnqueue_node_t));
	if(!new_node) return FAIL;

	/* set new node */
	new_node->next = tail;
	new_node->prev = tail->prev;

	/* alloc & copy dp reference to new_node */
	new_node->data = (pnqueue_data_t *)malloc(sizeof(pnqueue_data_t));
	if(!new_node->data) return FAIL;
	new_node->data->dp = data->dp;
	new_node->data->size = data->size;

	/* set tail & head */
	tail->prev->next = new_node;
	tail->prev = new_node;

	if(head->next == tail) {
		head->next = new_node;
	}

	handle->q_size++;

	return SUCC;
}

int pnqueue_get(pnqueue_handle_t *handle, pnqueue_data_t *data)
{
	pnqueue_node_t *front=NULL;
	pnqueue_node_t *next=NULL;
	pnqueue_node_t *head=NULL;
	pnqueue_node_t *tail=NULL;

	/* set temporarily */
	head = handle->head;
	tail = handle->tail;

	front = head->next;

	/* if front itself and next pointer indicate the tail, anymore none of node to get */
	if(front == tail && front->next == tail) return 0;

	next = front->next;

	head->next = next;
	next->prev = head;

	/* copy data */
	data->dp = front->data->dp;
	data->size = front->data->size;

	/* free momory */
	if(front->data) free(front->data);
	if(front) free(front);

	handle->q_size--;

	return 1;
}

void pnqueue_scan(pnqueue_handle_t *handle)
{
	handle->cur_p = handle->head;
}

pnqueue_node_t *pnqueue_next(pnqueue_handle_t *handle)
{
	handle->cur_p = handle->cur_p->next;

	if(handle->cur_p == handle->tail && handle->cur_p->next == handle->tail) return NULL;

	return handle->cur_p;
}

int pnqueue_get_size(pnqueue_handle_t *handle)
{
	return handle->q_size;
}
